@@ -51,7 +51,7 @@
 #define TRACKER_DB_MAX_FILE_SIZE      2000000000 
 
 /* Set current database version we are working with */
-#define TRACKER_DB_VERSION_NOW        TRACKER_DB_VERSION_4
+#define TRACKER_DB_VERSION_NOW        TRACKER_DB_VERSION_5
 #define TRACKER_DB_VERSION_FILE       "db-version.txt"
 
 typedef enum {
@@ -65,8 +65,8 @@ typedef enum {
 	TRACKER_DB_VERSION_1,       /* Version 0.6.6  (before indexer-split) */
 	TRACKER_DB_VERSION_2,       /* Version 0.6.90 (after  indexer-split) */
 	TRACKER_DB_VERSION_3,       /* Version 0.6.91 (stable release) */
-	TRACKER_DB_VERSION_4,       /* Version 0.6.92 (current TRUNK) */
-	TRACKER_DB_VERSION_5        /* Version 0.7    (vstore branch) */
+	TRACKER_DB_VERSION_4,       /* Version 0.6.92 */
+	TRACKER_DB_VERSION_5        /* Version 0.6.95 (with Composer tag) */
 } TrackerDBVersion;
 
 typedef struct {
@@ -2305,8 +2305,6 @@ tracker_db_manager_init (TrackerDBManagerFlags	flags,
 		 * new databases.
 		 */
 		tracker_ontology_shutdown ();
-
-		/* Make sure we initialize all other modules we depend on */
 		tracker_ontology_init ();
 
 		/* Now create the databases and close them */
@@ -2326,6 +2324,13 @@ tracker_db_manager_init (TrackerDBManagerFlags	flags,
 				dbs[i].iface = NULL;
 			}
 		}
+
+		/* Reset ontology - we do this so we repopulate our
+		 * ontology caches with the db_interface_create()
+		 * calls below.
+		 */
+		tracker_ontology_shutdown ();
+		tracker_ontology_init ();
 	} else {
 		if ((flags & TRACKER_DB_MANAGER_FORCE_NO_REINDEX) && need_reindex) {
 			g_message ("Reindex was needed, but has been forbidden by NO_REINDEX flag");
@@ -2341,7 +2346,11 @@ tracker_db_manager_init (TrackerDBManagerFlags	flags,
 			g_unlink (dbs[TRACKER_DB_CACHE].abs_filename);
 		}
 
-		/* Make sure we initialize all other modules we depend on */
+		/* In cases where we re-init this module, make sure
+		 * we have cleaned up the ontology before we load all
+		 * new databases.
+		 */
+		tracker_ontology_shutdown ();
 		tracker_ontology_init ();
 	}
 
