@@ -126,6 +126,11 @@ static INotifyHandle *libinotify_monitor_directory (TrackerMonitor *monitor,
 static void           libinotify_monitor_cancel    (gpointer        data);
 #endif /* USE_LIBINOTIFY */
 
+static void           tracker_monitor_update       (TrackerMonitor *monitor,
+						    const gchar    *module_name,
+						    GFile          *old_file,
+						    GFile          *new_file);
+
 static guint signals[LAST_SIGNAL] = { 0, };
 
 G_DEFINE_TYPE(TrackerMonitor, tracker_monitor, G_TYPE_OBJECT)
@@ -460,6 +465,7 @@ event_data_free (gpointer data)
 	event = data;
 	
 	g_object_unref (event->file);
+        g_free (event);
 }
 
 static guint
@@ -990,6 +996,7 @@ libinotify_monitor_force_emission (TrackerMonitor *monitor,
 		g_debug ("Cached event:%d being handled before %s",
 			 data->event_type,
 			 event_type_str);
+		g_free (event_type_str);
 
 		/* Signal event */
 		libinotify_cached_event_handle (monitor,
@@ -1226,6 +1233,11 @@ libinotify_monitor_event_cb (INotifyHandle *handle,
 				       other_file,
 				       is_directory, 
 				       TRUE);
+
+			if (is_directory) {
+				tracker_monitor_update (monitor, module_name, file, other_file);
+			}
+
 			g_hash_table_remove (monitor->private->event_pairs,
 					     GUINT_TO_POINTER (cookie));
 		}
@@ -1289,6 +1301,11 @@ libinotify_monitor_event_cb (INotifyHandle *handle,
 				       file,
 				       is_directory,
 				       is_source_indexed);
+
+			if (is_directory) {
+				tracker_monitor_update (monitor, module_name, other_file, file);
+			}
+
 			g_hash_table_remove (monitor->private->event_pairs,
 					     GUINT_TO_POINTER (cookie));
 		}
@@ -1663,6 +1680,39 @@ tracker_monitor_add (TrackerMonitor *monitor,
 	return TRUE;
 }
 
+static void
+tracker_monitor_update (TrackerMonitor *monitor,
+			const gchar    *module_name,
+			GFile          *old_file,
+			GFile          *new_file)
+{
+	GHashTable *monitors;
+	gpointer    file_monitor;
+
+	monitors = g_hash_table_lookup (monitor->private->modules, module_name);
+
+	if (!monitors) {
+		g_warning ("No monitor hash table for module:'%s'", module_name);
+		return;
+	}
+
+	file_monitor = g_hash_table_lookup (monitors, old_file);
+
+	if (!file_monitor) {
+		gchar *path;
+
+		path = g_file_get_path (old_file);
+		g_warning ("No monitor was found for directory:'%s'", path);
+		g_free (path);
+
+		return;
+	}
+
+	/* Replace key in monitors hashtable */
+	g_hash_table_steal (monitors, old_file);
+	g_hash_table_insert (monitors, new_file, file_monitor);
+}
+
 gboolean
 tracker_monitor_remove (TrackerMonitor *monitor,
 			const gchar    *module_name,
